import re
from pathlib import Path
from typing import Dict, Union

import pandas as pd
from sklearn.model_selection import train_test_split

from model_server.base.accessors import PatchStack, GenericImageFileAccessor
from model_server.base.roiset import PatchParams, RoiSet, RoiSetExportParams

class RoiSetIndex(dict):
    """
    Dictionary that identifies each unique RoiSet in the PhenoBase.
    """
    pass

class PhenoBase(object):
    def __init__(
            self,
            dataframe: pd.DataFrame,
            root: Path,
            info: Dict[str, str],
            roiset_subdir='phenobase',
            roiset_df_paths: list[Path] = None,
    ):
        """
        Create a database containing one or more RoiSets and their associated data
        :param dataframe: DataFrame object containing RoiSet information indexed by a unique RoiSet identifier
        :param root: top-level directory containing a directory with RoiSets
        :param info: about patch subdirectories, filters, etc.; passed to methods that create derivative PhenoBases
        :param roiset_subdir: name of subdirectory that contains serialized RoiSets
        :param roiset_df_paths: list of CSV files that individually define an Roiset
        """
        self.df = dataframe
        self.root = root
        self.info = info
        self.roiset_root_path = root / roiset_subdir
        self.roiset_df_paths = roiset_df_paths

        assert len(self.list_patch_series()) > 0, f'No patch collections found in {self.roiset_root_path}'
        for patch_series in self.list_patch_series():
            patches_path = self.roiset_root_path / patch_series
            self.verify_patches(
                self.df,
                patches_path.name,
                self.info['mask_patch_series'],
                self.roiset_root_path
            )

    def update(self):
        """
        Scan the target directory specified in .read() for additional entries and update accordingly.
        """
        list_df_paths = []
        for pa in (self.roiset_root_path / 'dataframe').iterdir():
            if pa.suffix.upper() == '.CSV' and pa not in self.roiset_df_paths:
                list_df_paths.append(pa)
                self.roiset_df_paths.append(pa)
        new_roisets = self._read_roiset_dfs(list_df_paths, max_stack_count=None, df_stacks=None)

        if len(new_roisets) == 0:
            return

        for patch_series in self.list_patch_series():
            patches_path = self.roiset_root_path / patch_series
            self.verify_patches(
                pd.concat(new_roisets),
                patches_path.name,
                self.info['mask_patch_series'],
                self.roiset_root_path
            )

        self.df = pd.concat([
            self.df,
            *new_roisets,
        ])

    def push(
            self,
            roiset: RoiSet,
            index_dict: RoiSetIndex,
            csv_prefix: str = None,
            patch_params: Dict[str, PatchParams] = {}
    ):
        """
        Push an RoiSet onto a PhenoBase based on a unique key
        :param roiset: RoiSet object to be added to PhenoBase
        :param index_dict: dict that uniquely identifies the RoiSet; keys must match existing ones
        :param csv_prefix:  (optional) prefix of newly created CSV file
        :param patch_params: (optional) dict of parameters for exporting each RoiSet patch series
        """
        # validate that index of RoiSet matches PhenoBase's schema and is unique
        index_str_parts = [f'{k}{v:04d}' for k, v in index_dict.items()]
        if csv_prefix is not None:
            index_str_parts.insert(0, csv_prefix)
        index_str = '-'.join(index_str_parts)
        for k in self.roiset_index:
            if k not in self.roiset_index:
                raise RoiSetIndexError(f'RoiSet index {k} not recognized in existing PhenoBase')
        for k, v in index_dict.items():
            if v in self.df.index.to_frame()[k]:
                raise RoiSetIndexError(f'RoiSet index {k}:{v} already existing in PhenoBase')

        # validate that patch sets are specified in RoiSet export parameters before attempting to export them
        if len(patch_params) == 0 and self.list_patch_series() is not None:
            raise MissingPatchSeriesError(f'RoiSet needs to specify patch export parameters: {self.list_patch_series()}')
        missing_patch_series = [
            psk for psk in self.list_patch_series() if psk.split('patches_')[-1] not in patch_params.keys()
        ]
        if any(missing_patch_series):
            raise MissingPatchSeriesError(f'Could not find patch series: {missing_patch_series}')

        # export RoiSet products and then re-scan target directory to update PhenoBase
        roiset.run_exports(
            self.roiset_root_path,
            prefix=index_str,
            params=RoiSetExportParams(
                patches=patch_params,
                write_patches_to_subdirectory=True,
            )
        )
        self.update()

    @staticmethod
    def _read_roiset_dfs(paths, max_stack_count, df_stacks=None):
        list_df = []
        for pa in paths[0: max_stack_count]:
            df_i = pd.read_csv(pa)
            roiset_index = {m[0]: int(m[1]) for m in re.findall(r'-*([a-zA-Z]+)(\d+)', pa.stem)}

            for k, v in roiset_index.items():
                df_i.insert(0, k, v)

            # transfer metadata from table of stack, if this is specified
            if df_stacks is not None:
                df_i['input_file'] = df_stacks['remote_path'].loc[roiset_index]

                # if ROIs are labeled the same in their input stacks, transfer this to ROI table
                if 'category_label' in df_stacks.columns:
                    df_i['category_label'] = df_stacks['category_label'].loc[roiset_index]

            # transfer in object classification results, if relevant
            annotations = [c for c in df_i.columns if c.startswith('classify_by')]
            if len(annotations) > 1 or 'category_label' in df_i.columns:
                raise AnnotationColumnsError('PhenoBase can only contain one annotation column')
            if len(annotations) == 1:
                df_i.rename(columns={annotations[0]: 'category_id'}, inplace=True)
                df_i['category_label'] = df_i['category_id']
            df_i.set_index([*roiset_index.keys(), 'label'], inplace=True)
            list_df.append(df_i)
        return list_df

    @classmethod
    def from_roiset(
            cls,
            root: Path,
            roiset: RoiSet,
            index_dict: RoiSetIndex,
            roiset_subdir='phenobase',
            patch_params: Dict[str, PatchParams] = {}
    ):
        # serialize RoiSet
        index_str = '-'.join([f'{k}{v:04d}' for k, v in index_dict.items()])
        roiset.run_exports(
            root / roiset_subdir,
            prefix=index_str,
            params=RoiSetExportParams(
                patches=patch_params,
                write_patches_to_subdirectory=True,
            )
        )
        return cls.read(root, roiset_subdir)

    @classmethod
    def read(
            cls,
            root: Path,
            roiset_subdir='phenobase',
            mask_patch_series: str = 'tight_patch_masks',
            max_stack_count: Union[int, None] = None,
            input_files_csv: str = None,
    ):
        """
        Automatically read serialized RoiSet from the specified directory and create a PhenoBase.  The resulting table
        parses its (generally multicolumn) index from RoiSet CSV filenames in the /dataframe subdirectory:
            roiset-x000-y001-t002 results in the index {x=0, y=1, t=2}
        and so on.
        :param root: top-level directory containing a directory with RoiSets
        :param roiset_subdir: name of subdirectory that contains serialized RoiSets
        :param mask_patch_series: name of subdirectory containing patch masks needed to deserialize each RoiSet
        :param max_stack_count: read only the specified number of RoiSets, or read all if None
        :param input_files_csv: (optional) the name of a file with supplementary information about RoiSets' input files
        :return: PhenoBase object
        """
        # concat single-stack dfs
        where_dfs = root / roiset_subdir / 'dataframe'
        roiset_df_paths = [pa for pa in where_dfs.iterdir() if pa.suffix.upper() == '.CSV']
        n_df = len(roiset_df_paths)
        assert n_df > 0, f'No patch dataframes found in {where_dfs}'
        if max_stack_count is None:
            max_stack_count = n_df

        if input_files_csv is not None:
            df_stacks = pd.read_csv(root / input_files_csv)
        else:
            df_stacks = None

        df_concat = pd.concat(
                cls._read_roiset_dfs(
                    roiset_df_paths,
                    max_stack_count=max_stack_count,
                    df_stacks=df_stacks,
                )
        )

        return cls(
            df_concat,
            root,
            {'mask_patch_series': mask_patch_series},
            roiset_df_paths=roiset_df_paths,
        )



    def list_patch_series(self):
        return [x.name for x in self.roiset_root_path.iterdir() if x.is_dir() and x.name.startswith('patches_')]

    @property
    def count(self):
        return len(self.df)

    def chroma(self, patch_series: str):
        return self.sample(1).get_raw_patchstack(patch_series).chroma

    def filter(self, filters) -> pd.DataFrame:
        if filters is None:
            self.info['filters'] = None
            return self
        else: # parse filters
            query_str = 'label > 0'
            for k, val in filters.items():
                assert k in ('area', 'diag', 'min_hw', 'diag')
                if val is None:
                    continue
                vmin = val['min']
                vmax = val['max']
                assert vmin >= 0
                query_str = query_str + f' & {k} > {vmin} & {k} < {vmax}'
                info = {'filters': query_str, **self.info}
            return self.__class__(self.df.query(query_str), self.root, info)

    def sample(self, n_max: int):
        if n_max is None or n_max >= self.count:
            return self
        else:
            return self.__class__(self.df.sample(int(n_max)), self.root, self.info)

    def get_rows(self, rows: pd.Index) -> pd.DataFrame:
        return self.__class__(self.df.loc[rows], self.root, self.info)

    def write_df(self):
        return self.df.to_csv(self.root / 'phenobase.csv')

    def get_raw_patchstack(self, patch_series: str) -> PatchStack:
        def _get_patch(roi):
            pa = getattr(roi, f'{patch_series}_path')
            fp = self.roiset_root_path / pa
            acc = GenericImageFileAccessor.read(fp)
            return acc.data
        return PatchStack([_get_patch(r) for r in self.df.itertuples()], force_ydim_longest=True)

    def get_patch_masks(self) -> PatchStack:
        patch_mask_column = self.info['mask_patch_series'] + '_path'
        def _get_mask(roi):
            fp = self.roiset_root_path / getattr(roi, patch_mask_column)
            return GenericImageFileAccessor.read(fp).data
        return PatchStack([_get_mask(r) for r in self.df.itertuples()], force_ydim_longest=True)

    def get_patch_obmaps(self) -> PatchStack:
        if 'category_id' not in self.df.columns:
            raise AnnotationColumnsError(
                'PhenoBase does not have an annotation column with which to generate object maps'
            )
        patch_mask_column = self.info['mask_patch_series'] + '_path'
        def _get_obmap(roi):
            fp = self.roiset_root_path / getattr(roi, patch_mask_column)
            nda = GenericImageFileAccessor.read(fp).data
            return ((nda > 0) * roi.category_id).astype('uint8')
        return PatchStack([_get_obmap(r) for r in self.df.itertuples()], force_ydim_longest=True)

    @staticmethod
    def verify_patches(df, patch_series: str, mask_patch_series: str, root: PatchStack):
        # verify patch files and put paths in column
        all_exist = {}
        for k in [f'{patch_series}_path', f'{mask_patch_series}_path']:
            df[f'{k}_exists'] = df[k].apply(lambda x: (root / x).exists())
            all_exist[k] = df[f'{k}_exists'].all()
        assert all(all_exist), f'Could not verify that all patches in {patch_series} exist'

    @property
    def labels(self):
        if 'category_id' not in self.df.columns and 'category_label' not in self.df.columns:
            return None
        labels_dict = {g[0][0]: g[0][1] for g in self.df.groupby(['category_id', 'category_label'])}
        return [labels_dict.get(i, 'undefined') for i in range(1, 1 + max(labels_dict.keys()))]

    def split(self, test_size=0.3) -> Dict[str, pd.Series]:
        tr_rows, te_rows = train_test_split(self.df.index, test_size=test_size)

        return {
            'train': self.__class__(self.df.loc[tr_rows], self.root, {'split': 'train', **self.info}),
            'test': self.__class__(self.df.loc[te_rows], self.root, {'split': 'test', **self.info})
        }

    @property
    def roiset_index(self) -> list:
        return [n for n in self.df.index.names if n.upper() != 'LABEL']


class Error(Exception):
    pass

class InvalidPatchStackError(Error):
    pass

class RoiSetIndexError(Error):
    pass

class AnnotationColumnsError(Error):
    pass

class MissingPatchSeriesError(Error):
    pass
